/*++========================================================================
 * Program Name:     Novell NCP Redirector for Linux
 * File Name:        daemon.c
 * Version:          v1.00
 * Author:           James Turner
 *
 * Abstract:         This module contains all the functions necessary
 *                   for sending commands to our daemon module.
 * Notes:
 * Revision History: 
 *
 *
 * Copyright (C) 2005 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *======================================================================--*/

/*===[ Include files specific to Linux ]==================================*/
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/poll.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>


/*===[ Include files specific to this module ]============================*/
#include "vfs.h"
#include "nwcapi.h"
#include "commands.h"
#include "nwerror.h"

/*===[ External data ]====================================================*/
extern char *Novfs_CurrentMount;

/*===[ External prototypes ]==============================================*/
extern int DbgPrint( char *Fmt, ... );
extern void mydump(int size, void *dumpptr);

extern char *Scope_Get_UserName( void *Scope );
extern void *Scope_Lookup( void );
extern session_t Scope_Get_SessionId( void *Scope );
extern void Scope_Cleanup( void );

extern int Novfs_Read_Stream( u_long ConnHandle, u_char *Handle, u_char *Buffer, size_t *Bytes, loff_t *Offset, int User, session_t SessionId);
extern int Novfs_Write_Stream( u_long ConnHandle, u_char *Handle, u_char *Buffer, size_t *Bytes, loff_t *Offset, session_t SessionId);
extern int Novfs_Close_Stream( u_long ConnHandle, u_char *Handle, session_t SessionId );

extern int Novfs_Add_to_Root(char *);

/*
 * nwcapi.c functions
 */
extern int NwAuthConnWithId(PXPLAT pdata, session_t Session);
extern int NwConnClose(PXPLAT pdata, u_long *Handle, session_t Session);
extern int NwGetConnInfo(PXPLAT pdata, session_t Session);
extern int NwSetConnInfo(PXPLAT pdata, session_t Session);
extern int NwGetDaemonVersion(PXPLAT pdata, session_t Session);
extern int NwGetIdentityInfo(PXPLAT pdata, session_t Session);
extern int NwLicenseConn(PXPLAT pdata, session_t Session);
extern int NwLoginIdentity(PXPLAT pdata, session_t Session);
extern int NwLogoutIdentity(PXPLAT pdata, session_t Session);
extern int NwOpenConnByAddr(PXPLAT pdata, u_long *Handle, session_t Session);
extern int NwOpenConnByName(PXPLAT pdata, u_long *Handle, session_t Session);
extern int NwOpenConnByRef(PXPLAT pdata, u_long *Handle, session_t Session);
extern int NwQueryFeature(PXPLAT pdata, session_t Session);
extern int NwRawSend(PXPLAT pdata, session_t Session);
extern int NwScanConnInfo(PXPLAT pdata, session_t Session);
extern int NwSysConnClose(PXPLAT pdata, u_long *Handle, session_t Session);
extern int NwUnAuthenticate(PXPLAT pdata, session_t Session);
extern int NwUnlicenseConn(PXPLAT pdata, session_t Session);
extern int NwcChangeAuthKey(PXPLAT pdata, session_t Session);
extern int NwcEnumIdentities(PXPLAT pdata, session_t Session);
extern int NwcGetDefaultNameCtx(PXPLAT pdata, session_t Session);
extern int NwcGetPreferredDSTree(PXPLAT pdata, session_t Session);
extern int NwcGetTreeMonitoredConn(PXPLAT pdata, session_t Session);
extern int NwcSetDefaultNameCtx(PXPLAT pdata, session_t Session);
extern int NwcSetPreferredDSTree(PXPLAT pdata, session_t Session);
extern int NwcSetPrimaryConn(PXPLAT pdata, session_t Session);
extern int NwcGetPrimaryConn(PXPLAT pdata, session_t Session);
extern int NwcSetMapDrive(PXPLAT pdata, session_t Session);
extern int NwcUnMapDrive(PXPLAT pdata, session_t Session);
extern int NwcEnumerateDrives(PXPLAT pdata, session_t Session);
extern int NwcGetBroadcastMessage(PXPLAT pdata, session_t Session);
extern int NwdSetKeyValue(PXPLAT pdata, session_t Session);
extern int NwdVerifyKeyValue(PXPLAT pdata, session_t Session);

/*===[ Manifest constants ]===============================================*/
#define QUEUE_SENDING 0
#define QUEUE_WAITING 1
#define QUEUE_TIMEOUT 2
#define QUEUE_ACKED   3
#define QUEUE_DONE    4

#define TIMEOUT_VALUE 10

#define DH_TYPE_UNDEFINED    0
#define DH_TYPE_STREAM       1
#define DH_TYPE_CONNECTION   2

/*===[ Type definitions ]=================================================*/
typedef struct _DAEMON_QUEUE
{
   struct list_head list;       /* Must be first entry */
   spinlock_t       lock;       /* Used to control access to list */
   struct semaphore semaphore;  /* Used to signal when data is available */
} daemon_queue_t;

typedef struct _DAEMON_COMMAND
{
   struct list_head  list;      /* Must be first entry */
   atomic_t          reference;
   u_long            status;
   u_long            flags;
   struct semaphore  semaphore;
   u_long            sequence;
   struct timer_list timer;
   void              *request;
   u_long            reqlen;
   void              *data;
   int               datalen;
   void              *reply;
   u_long            replen;
}  daemon_command_t;


typedef struct _DAEMON_HANDLE_
{
   struct list_head  list;
   rwlock_t          lock;
   session_t         session;
} daemon_handle_t;

typedef struct _DAEMON_RESOURCE_
{
   struct list_head  list;
   int               type;
   u_long            connection;
   u_char            handle[6];
   mode_t            mode; 
   loff_t            size;
} daemon_resource_t;

/*===[ Function prototypes ]==============================================*/
int     Daemon_Added_Resource(daemon_handle_t *DHandle, int Type, u_long CHandle, u_char *FHandle, u_long Mode, u_long Size);
int     Daemon_Close_Control(struct inode *Inode, struct file *File);
int     Daemon_CreateSessionId( uint64_t *SessionId );
int     Daemon_DestroySessionId( uint64_t SessionId );
void    Daemon_Dumpque( void );
int     Daemon_Get_UserSpace( uint64_t SessionId,  uint64_t *TotalSize, uint64_t *TotalFree, uint64_t *TotalDirectoryEnties, uint64_t *FreeDirectoryEnties);
int     Daemon_Library_close(struct inode *inode, struct file *file);
int     Daemon_Library_ioctl(struct inode *inode, struct file *file, unsigned int cmd, u_long arg);
int     Daemon_Library_open(struct inode *inode, struct file *file);
ssize_t Daemon_Library_read(struct file *file, char *buf, size_t len, loff_t *off);
ssize_t Daemon_Library_write(struct file *file, const char *buf, size_t len, loff_t *off);
int     Daemon_Login(PLOGIN Login, session_t Session);
int     Daemon_Logout(PLOGOUT Logout, session_t Session);
int     Daemon_Open_Control(struct inode *Inode, struct file *File);
uint    Daemon_Poll(struct file *file, struct poll_table_struct *poll_table);
ssize_t Daemon_Receive_Reply(struct file * file, char * buf, size_t nbytes, loff_t *ppos);
int     Daemon_Remove_Resource(daemon_handle_t *DHandle, int Type, u_long CHandle, u_long FHandle);

ssize_t Daemon_Send_Command(struct file *file, char *buf, size_t len, loff_t *off);
int     Daemon_SetMountPoint( char *Path );
void    Daemon_Timer(u_long data);
int     Daemon_getpwuid( uid_t uid, int unamelen, char *uname );
int     Daemon_ioctl(struct inode *inode, struct file *file, unsigned int cmd, u_long arg);
void    Init_Daemon_Queue( void );
int     Queue_Daemon_Command(void *request, u_long reqlen, void *data, int dlen, void **reply, u_long *replen, int interruptible);
void    Queue_get(daemon_command_t *que);
void    Queue_put(daemon_command_t *que);
void    Uninit_Daemon_Queue( void );
int     do_login(NclString *Server, NclString *Username, NclString *Password, u_long *lgnId, session_t Session);
int     do_logout( struct qstr *Server, session_t Session );
daemon_command_t *find_queue(u_long sequence);
daemon_command_t *get_next_queue(int Set_Queue_Waiting);
int NwdConvertNetwareHandle(PXPLAT pdata, daemon_handle_t *DHandle);
int NwdConvertLocalHandle(PXPLAT pdata, daemon_handle_t *DHandle);
int NwdGetMountPath(PXPLAT pdata);

/*===[ Global variables ]=================================================*/
daemon_queue_t Daemon_Queue;

static DECLARE_WAIT_QUEUE_HEAD(Read_waitqueue);

u_long Sequence = 0;
atomic_t Daemon_Open_Count=ATOMIC_INIT(0);

u_long Daemon_Command_Timeout = TIMEOUT_VALUE;

/*===[ Code ]=============================================================*/

/*++======================================================================*/
void Init_Daemon_Queue()
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   INIT_LIST_HEAD(&Daemon_Queue.list);
   spin_lock_init(&Daemon_Queue.lock);
   init_MUTEX_LOCKED(&Daemon_Queue.semaphore);
}

/*++======================================================================*/
void Uninit_Daemon_Queue( void )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   /* Does nothing for now but we maybe should clear the queue. */
}

/*++======================================================================*/
void 
NO_TRACE
Daemon_Timer(u_long data)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que = (daemon_command_t *)data;

   if ( QUEUE_ACKED != que->status )
   {
      que->status = QUEUE_TIMEOUT;
   }
   up(&que->semaphore);
}

/*++======================================================================*/
int Queue_Daemon_Command(
   void *request,
   u_long reqlen, 
   void *data, 
   int dlen,
   void **reply, 
   u_long *replen,
   int interruptible)
/*
 *
 *  Arguments:     void *request - pointer to the request that is to be sent.  Needs to be kernel memory.
 *                 int reqlen - length of the request.
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;
   int retCode = 0;
   u_int64_t ts1, ts2;

   rdtscll(ts1);

   DbgPrint( "Queue_Daemon_Command: 0x%x %d\n", request, reqlen);

   if (atomic_read(&Daemon_Open_Count))
   {

      que = (daemon_command_t *)Novfs_Malloc(sizeof(*que), GFP_KERNEL);

      DbgPrint( "Queue_Daemon_Command: que=0x%x\n", que);
      if (que)
      {
         atomic_set( &que->reference, 0 );
         que->status = QUEUE_SENDING;
         que->flags  = 0;

         init_MUTEX_LOCKED(&que->semaphore);

         DbgPrint( "Queue_Daemon_Command: semaphore inited que=0x%x\n", que);
         
         que->sequence = InterlockedIncrement(&Sequence);

         DbgPrint( "Queue_Daemon_Command: sequence=0x%x\n", que->sequence);

         ((PCOMMAND_REQUEST_HEADER)request)->SequenceNumber = que->sequence;

         /*
          * Setup and start que timer
          */
         init_timer(&que->timer);
         que->timer.expires = jiffies + (HZ * Daemon_Command_Timeout);
         que->timer.data = (u_long)que;
         que->timer.function = Daemon_Timer;
         add_timer(&que->timer);

         DbgPrint( "Queue_Daemon_Command: timer started que=0x%x\n", que);

         /*
          * Setup request
          */
         que->request = request;
         que->reqlen  = reqlen;
         que->data    = data;
         que->datalen = dlen;
         que->reply   = NULL;
         que->replen  = 0;

         DbgPrint( "Queue_Daemon_Command: setting up que=0x%x\n", que);

         /*
          * Added entry to queue.
          */
         DbgPrint( "Queue_Daemon_Command: Daemon_Queue locked\n");

         /*
          * Check to see if interruptible and set flags.
          */
         if (interruptible) 
         {
            que->flags |= INTERRUPTIBLE;
         }

         spin_lock(&Daemon_Queue.lock);
         Queue_get( que );
         list_add_tail(&que->list, &Daemon_Queue.list);
         spin_unlock(&Daemon_Queue.lock);

         DbgPrint( "Queue_Daemon_Command: 0x%x added to Daemon_Queue\n", que);
         /*
          * Signal that there is data to be read
          */
         up(&Daemon_Queue.semaphore);

         /*
          * Give a change to the other processes.
          */
         yield();

         DbgPrint( "Queue_Daemon_Command: calling down 0x%x\n", CURRENT_TIME);

         /*
          * Block waiting for reply or timeout
          */
         down(&que->semaphore);

         if ( QUEUE_ACKED == que->status )
         {
            que->status = QUEUE_WAITING;
            mod_timer(&que->timer, jiffies + (HZ * 2 * Daemon_Command_Timeout));
            if (interruptible) 
            {
               retCode = down_interruptible(&que->semaphore);
            }
            else
            {
               down(&que->semaphore);
            }
         }

         DbgPrint( "Queue_Daemon_Command: after down 0x%d 0x%x\n", retCode, CURRENT_TIME);
         /*
          * Delete timer
          */
         del_timer(&que->timer);

         /*
          * Check for timeout
          */
         if ((QUEUE_TIMEOUT == que->status) && (NULL == que->reply) )
         {
            DbgPrint( "Queue_Daemon_Command: Timeout\n");
            retCode = -ETIME;
         }
         *reply  = que->reply;
         *replen = que->replen;

         /*
          * Remove item from queue
          */
         Queue_put( que );

      }
      else /* Error case with no memory */
      {
         retCode = -ENOMEM;
         *reply = NULL;
         *replen = 0;
      }
   }
   else
   {
      retCode = -EIO;
      *reply = NULL;
      *replen = 0;

   }
   rdtscll(ts2);
   ts2 = ts2-ts1;

   DbgPrint( "Queue_Daemon_Command: %llu retCode=%d \n", ts2, retCode);
   return(retCode);
}

/*++======================================================================*/
void Queue_get(daemon_command_t *Que)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   DbgPrint("Queue_get: que=0x%p %d\n", Que, atomic_read(&Que->reference));
   atomic_inc( &Que->reference );
}

/*++======================================================================*/
void Queue_put(daemon_command_t *Que)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   spin_lock(&Daemon_Queue.lock);

   DbgPrint("Queue_put: que=0x%p %d\n", Que, atomic_read(&Que->reference));
   if ( atomic_dec_and_test( &Que->reference ))
   {
      /*
       * Remove item from queue
       */
      list_del(&Que->list);

      spin_unlock(&Daemon_Queue.lock);

      /*
       * Free item memory
       */
      Novfs_Free(Que);
   }
   else
   {
      spin_unlock(&Daemon_Queue.lock);
   }
}

/*++======================================================================*/
daemon_command_t *get_next_queue(int Set_Queue_Waiting)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;
   
   spin_lock(&Daemon_Queue.lock);
   que = (daemon_command_t *)Daemon_Queue.list.next;

   DbgPrint( "get_next_queue: que=0x%x\n", que);

   while (que && (que != (daemon_command_t *)&Daemon_Queue.list.next) && ( que->status != QUEUE_SENDING ) )
   {
      que = (daemon_command_t *)que->list.next;
      DbgPrint( "get_next_queue: que1=0x%x\n", que);
   }

   if ((NULL == que) || (que == (daemon_command_t *)&Daemon_Queue.list) || (que->status != QUEUE_SENDING))
   {
      que = NULL;
   }
   else if (Set_Queue_Waiting)
   {
      que->status = QUEUE_WAITING;
   }

   if (que)
   {
      atomic_inc( &que->reference );
   }

   spin_unlock(&Daemon_Queue.lock);

   DbgPrint( "get_next_queue: return=0x%x\n", que);
   return(que);
}

/*++======================================================================*/
daemon_command_t *find_queue(u_long sequence)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;
   
   DbgPrint( "find_queue: 0x%x\n", sequence);

   spin_lock(&Daemon_Queue.lock);
   que = (daemon_command_t *)Daemon_Queue.list.next;
   
   DbgPrint( "find_queue: que=0x%x\n", que);

   while (que && (que != (daemon_command_t *)&Daemon_Queue.list.next) && (que->sequence != sequence))
   {
      DbgPrint( "find_queue: que1=0x%x\n", que);
      que = (daemon_command_t *)que->list.next;
   }

   if ((NULL == que) || (que == (daemon_command_t *)&Daemon_Queue.list.next) || (que->sequence != sequence))
   {
      que = NULL;
   }

   if (que)
   {
      atomic_inc( &que->reference );
   }

   spin_unlock(&Daemon_Queue.lock);

   DbgPrint( "find_queue: return 0x%x\n", que);
   return(que);
}
/*++======================================================================*/
int Daemon_Open_Control(struct inode *Inode, struct file *File)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   DbgPrint("Daemon_Open_Control: pid=%d Count=%d\n", current->pid, atomic_read(&Daemon_Open_Count));
   atomic_inc(&Daemon_Open_Count);

   return(0);
}

/*++======================================================================*/
int Daemon_Close_Control(struct inode *Inode, struct file *File)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;

   DbgPrint("Daemon_Close_Control: pid=%d Count=%d\n", current->pid, atomic_read(&Daemon_Open_Count));

   if (atomic_dec_and_test(&Daemon_Open_Count))
   {
      /*
       * Signal any pending que itmes.
       */

      spin_lock(&Daemon_Queue.lock);
      que = (daemon_command_t *)Daemon_Queue.list.next;

      DbgPrint( "Daemon_Close_Control: que=0x%x\n", que);

      while (que && (que != (daemon_command_t *)&Daemon_Queue.list.next) && (que->status != QUEUE_DONE))
      {
         que->status = QUEUE_TIMEOUT;
         up(&que->semaphore);

         que = (daemon_command_t *)que->list.next;
         DbgPrint( "Daemon_Close_Control: que1=0x%x\n", que);
      }
      spin_unlock(&Daemon_Queue.lock);

      Scope_Cleanup();
   }

   return(0);
}

/*++======================================================================*/
ssize_t 
Daemon_Send_Command(struct file *file, char *buf, size_t len, loff_t *off)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;
   size_t retValue = 0;
   int Finished=0;
   PDATA_LIST dlist;
   int i, dcnt, bcnt, ccnt, error;
   char *vadr;

   DbgPrint( "Daemon_Send_Command: %u %lld\n", len, *off);

   while ( !Finished )
   {
      que = get_next_queue(1);
      DbgPrint( "Daemon_Send_Command: 0x%x\n", que);
      if (que)
      {
         retValue = que->reqlen;
         if (retValue > len)
         {
            retValue = len;
         }
         if (retValue > 0x80)
            mydump(0x80, que->request);
         else
            mydump(retValue, que->request);

         copy_to_user(buf, que->request, retValue);
         if (que->datalen && (retValue < len))
         {
            buf += retValue;
            dlist = que->data;
            dcnt  = que->datalen;
            for (i=0; i<dcnt; i++, dlist++)
            {
               if ( DLREAD == dlist->rwflag )
               {
                  bcnt = dlist->len;
                  DbgPrint("Daemon_Send_Command%d: page=0x%x offset=0x%x len=%d\n", i, dlist->page, dlist->offset, dlist->len);
                  if ((bcnt + retValue) <= len)
                  {
                     if (dlist->page)
                     {
                        vadr = kmap(dlist->page);
                        vadr += (u_int)dlist->offset;
                     }
                     else
                     {
                        vadr = dlist->offset;
                     }
                     ccnt = copy_to_user(buf, vadr, bcnt);
                     if (dlist->page)
                     {
                        kunmap(dlist->page);
                     }

                     DbgPrint("Daemon_Send_Command: Copy %d from 0x%x to 0x%x.\n", bcnt, vadr, buf);
                     if (bcnt > 0x80)
                        mydump(0x80, vadr);
                     else
                        mydump(bcnt, vadr);

                     retValue += bcnt;
                     buf += bcnt;
                  }
                  else
                  {
                     break;
                  }
               }
            }
         }
         Queue_put( que );
         break;
      }

      if (O_NONBLOCK & file->f_flags)
      {
         retValue = -EAGAIN;
         break;
      }
      else
      {
         if ((error = down_interruptible(&Daemon_Queue.semaphore)))
         {
            DbgPrint( "Daemon_Send_Command: after down_interruptible error...%d\n", error);
            retValue = -EINTR;
            break;
         }
         DbgPrint( "Daemon_Send_Command: after down_interruptible\n");
      }
   }

   *off = *off;

   DbgPrint( "Daemon_Send_Command: return 0x%x\n", retValue);

   return(retValue);
}

/*++======================================================================*/
ssize_t
Daemon_Receive_Reply(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;
   size_t retValue = 0;
   void *reply;
   u_long sequence;

   PDATA_LIST dlist;
   char *vadr;
   int i;

   DbgPrint( "Daemon_Receive_Reply: buf=0x%x nbytes=%d ppos=%llx\n", buf, nbytes, *ppos);

   /*
    * Get sequence number from reply buffer
    */

   copy_from_user(&sequence, buf, sizeof(sequence));

   /*
    * Find item based on sequence number
    */
   que = find_queue(sequence);

   DbgPrint( "Daemon_Receive_Reply: 0x%x 0x%x %d\n", sequence, que, nbytes);
   if (que)
   {
      do
      {
         retValue = nbytes;
         /*
          * Ack packet from novfsd.  Remove timer and
          * return
          */
         if (nbytes == sizeof(sequence))
         {
            que->status = QUEUE_ACKED;
            break;
         }

         /*
          * Set status that packet is done.
          */
         que->status = QUEUE_DONE;

         if ( NULL != (dlist = que->data) )
         {
            int thiscopy, left=nbytes;
            retValue = 0;


            DbgPrint( "Daemon_Receive_Reply: dlist=0x%x count=%d\n", dlist, que->datalen);
            for (i=0; i < que->datalen; i++)
            {
               DbgPrint( "Daemon_Receive_Reply:\n" \
                         "   dlist[%d].page:   0x%x\n" \
                         "   dlist[%d].offset: 0x%x\n" \
                         "   dlist[%d].len:    0x%x\n" \
                         "   dlist[%d].rwflag: 0x%x\n",
                  i, dlist[i].page,
                  i, dlist[i].offset,
                  i, dlist[i].len,
                  i, dlist[i].rwflag);

               if (DLWRITE == dlist[i].rwflag)
               {
                  vadr = NULL;
                  if (dlist[i].page)
                  {
                     vadr = kmap(dlist[i].page);
                     vadr += (u_int)dlist[i].offset;
                  }
                  else
                  {
                     vadr = dlist[i].offset;
                  }
                  thiscopy = dlist[i].len;
                  if (thiscopy > left)
                  {
                     thiscopy = left;
                     dlist[i].len = left;
                  }
                  copy_from_user(vadr, buf, thiscopy);
                  if (thiscopy > 0x80)
                     mydump(0x80, vadr);
                  else
                     mydump(thiscopy, vadr);
                  left -= thiscopy;
                  retValue += thiscopy;
                  buf += thiscopy;
               }
            }
         }
         else
         {
            reply = Novfs_Malloc(nbytes, GFP_KERNEL);
            DbgPrint( "Daemon_Receive_Reply: reply=0x%x\n", reply);
            if (reply)
            {
               retValue = nbytes;
               que->reply = reply;
               que->replen = nbytes;

               retValue -= copy_from_user(reply, buf, retValue);
               if (retValue > 0x80)
                  mydump(0x80, reply);
               else
                  mydump(retValue, reply);


            }
            else
            {
               retValue = -ENOMEM;
            }
         }
         up(&que->semaphore);
      } while (0);
      Queue_put( que );
   } 

   DbgPrint( "Daemon_Receive_Reply: return 0x%x\n", retValue);

   return(retValue);
}

/*++======================================================================*/
int do_login(NclString *Server, NclString *Username, NclString *Password, u_long *lgnId, session_t Session)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   PLOGIN_USER_REQUEST cmd;
   PLOGIN_USER_REPLY   reply;
   u_long replylen=0;
   int retCode, cmdlen, datalen;
   u_char *data;

   datalen = Server->len+Username->len+Password->len;
   cmdlen = sizeof(*cmd) + datalen;
   cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
   if (cmd)
   {
      data = (u_char *)cmd + sizeof(*cmd);
      cmd->Command.CommandType    = VFS_COMMAND_LOGIN_USER;
      cmd->Command.SequenceNumber = 0;
      cmd->Command.SessionId      = Session;

      cmd->srvNameType    = Server->type;
      cmd->serverLength   = Server->len;
      cmd->serverOffset   = (u_long)(data-(u_char *)cmd);
      memcpy(data, Server->buffer, Server->len);
      data += Server->len;

      cmd->usrNameType    = Username->type;
      cmd->userNameLength = Username->len;
      cmd->userNameOffset = (u_long)(data-(u_char *)cmd);
      memcpy(data, Username->buffer, Username->len);
      data += Username->len;

      cmd->pwdNameType    = Password->type;
      cmd->passwordLength = Password->len;
      cmd->passwordOffset = (u_long)(data-(u_char *)cmd);
      memcpy(data, Password->buffer, Password->len);
      data += Password->len;

      retCode = Queue_Daemon_Command(cmd, cmdlen, NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
      if (reply)
      {
         if (reply->Reply.ErrorCode)
         {
            retCode = reply->Reply.ErrorCode;
         }
         else
         {
            retCode = 0;
            if (lgnId)
            {
               *lgnId = reply->loginIdentity;
            }
         }
         Novfs_Free(reply);
      }
      memset(cmd, 0, cmdlen);
      Novfs_Free(cmd);
   }
   else
   {
      retCode = -ENOMEM;
   }
   return( retCode );

}

/*++======================================================================*/
int do_logout( struct qstr *Server, session_t Session )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   PLOGOUT_REQUEST cmd;
   PLOGOUT_REPLY   reply;
   u_long replylen=0;
   int retCode, cmdlen;

   cmdlen = (int)(&((PLOGOUT_REQUEST)0)->Name) + Server->len;
   cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
   if (cmd)
   {
      cmd->Command.CommandType    = VFS_COMMAND_LOGOUT_USER;
      cmd->Command.SequenceNumber = 0;
      cmd->Command.SessionId      = Session;

      cmd->Length   = Server->len;
      memcpy(cmd->Name, Server->name, Server->len);

      retCode = Queue_Daemon_Command(cmd, cmdlen, NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
      if (reply)
      {
         if (reply->Reply.ErrorCode)
         {
            retCode = -EIO;
         }
         Novfs_Free(reply);
      }
      Novfs_Free(cmd);
   }
   else
   {
      retCode = -ENOMEM;
   }
   return( retCode );

}

/*++======================================================================*/
int Daemon_getpwuid( uid_t uid, int unamelen, char *uname )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   GETPWUID_REQUEST cmd;
   PGETPWUID_REPLY  reply;
   u_long replylen=0;
   int retCode;

   cmd.Command.CommandType    = VFS_COMMAND_GETPWUD;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId      = 0;
   cmd.uid = uid;

   retCode = Queue_Daemon_Command(&cmd, sizeof(cmd), NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
   if (reply)
   {
      if (reply->Reply.ErrorCode)
      {
         retCode = -EIO;
      }
      else
      {
         retCode = 0;
         memset(uname, 0, unamelen);
         replylen = replylen - (int)(&((PGETPWUID_REPLY)0)->UserName);
         if (replylen)
         {
            if (replylen > unamelen)
            {
               retCode = -EINVAL;
               replylen = unamelen-1;
            }
            memcpy(uname, reply->UserName, replylen);
         }
      }
      Novfs_Free(reply);
   }
   return( retCode );

}

/*++======================================================================*/
int Daemon_getversion( char *Buf, int Length )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   GET_VERSION_REQUEST cmd;
   PGET_VERSION_REPLY  reply;
   u_long replylen=0;
   int retVal=0;

   cmd.Command.CommandType    = VFS_COMMAND_GET_VERSION;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId      = 0;

   Queue_Daemon_Command(&cmd, sizeof(cmd), NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
   if (reply)
   {
      if (reply->Reply.ErrorCode)
      {
         retVal = -EIO;
      }
      else
      {
         retVal = replylen - (int)(&((PGET_VERSION_REPLY)0)->Version);
         if (retVal < Length)
         {
            memcpy(Buf, reply->Version, retVal);
            Buf[retVal] = '\0';
         }
      }
      Novfs_Free(reply);
   }
   return( retVal );

}

/*++======================================================================*/
int Daemon_Login(PLOGIN Login, session_t Session)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -ENOMEM;
   LOGIN lLogin;
   NclString   server;
   NclString   username;
   NclString   password;

   if ( !copy_from_user(&lLogin, Login, sizeof(lLogin)))
   {
      if ( (server.buffer = Novfs_Malloc(lLogin.Server.Length, GFP_KERNEL)) )
      {
         server.len = lLogin.Server.Length;
         server.type = NWC_STRING_TYPE_ASCII;
         if ( !copy_from_user((void *)server.buffer, lLogin.Server.Data, server.len) )
         {
            if ( (username.buffer = Novfs_Malloc(lLogin.UserName.Length, GFP_KERNEL)) )
            {
               username.len = lLogin.UserName.Length;
               username.type = NWC_STRING_TYPE_ASCII;
               if ( !copy_from_user((void *)username.buffer, lLogin.UserName.Data, username.len) )
               {
                  if ( (password.buffer = Novfs_Malloc(lLogin.Password.Length, GFP_KERNEL)) )
                  {
                     password.len = lLogin.Password.Length;
                     password.type = NWC_STRING_TYPE_ASCII;
                     if ( !copy_from_user((void *)password.buffer, lLogin.Password.Data, password.len) )
                     {
                        retCode = do_login(&server, &username, &password, NULL, Session);
                        if ( !retCode )
                        {
                           char *username;
                           username = Scope_Get_UserName( NULL );
                           if (username)
                           {
                              Novfs_Add_to_Root( username );
                           }
                        }
                     }
                     memset(password.buffer, 0, password.len);
                     Novfs_Free(password.buffer);
                  }
               }
               memset(username.buffer, 0, username.len);
               Novfs_Free(username.buffer);
            }
         }
         Novfs_Free(server.buffer);
      }
   }

   return( retCode );
}

/*++======================================================================*/
int Daemon_Logout(PLOGOUT Logout, session_t Session)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   LOGOUT lLogout;
   struct qstr server;
   int retCode=0;

   if ( !copy_from_user(&lLogout, Logout, sizeof(lLogout)))
   {
      if ( (server.name = Novfs_Malloc(lLogout.Server.Length, GFP_KERNEL)) )
      {
         server.len = lLogout.Server.Length;
         if ( !copy_from_user((void *)server.name, lLogout.Server.Data, server.len) )
         {
            retCode = do_logout( &server, Session );
         }
         Novfs_Free(server.name);
      }
   }
   return( retCode );
}

/*++======================================================================*/
int Daemon_CreateSessionId( uint64_t *SessionId )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   CREATE_CONTEXT_REQUEST cmd;
   PCREATE_CONTEXT_REPLY  reply;
   u_long replylen=0;
   int retCode=0;

   DbgPrint("Daemon_CreateSessionId: %d\n", current->pid);

   cmd.Command.CommandType = VFS_COMMAND_CREATE_CONTEXT;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId = 0;

   retCode = Queue_Daemon_Command(&cmd, sizeof(cmd), NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
   if ( reply )
   {
      if ( !reply->Reply.ErrorCode  && replylen > sizeof(COMMAND_REPLY_HEADER))
      {
         *SessionId = reply->SessionId;
         retCode = 0;
      }
      else
      {
         *SessionId = 0;
         retCode = -EIO;
      }
      Novfs_Free(reply);
   }
   DbgPrint("Daemon_CreateSessionId: SessionId=0x%llx\n", *SessionId);
   return(retCode);
}

/*++======================================================================*/
int Daemon_DestroySessionId( uint64_t SessionId )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   DESTROY_CONTEXT_REQUEST cmd;
   PDESTROY_CONTEXT_REPLY  reply;
   u_long replylen=0;
   int retCode=0;

   DbgPrint("Daemon_DestroySessionId: 0x%llx\n", SessionId);

   cmd.Command.CommandType = VFS_COMMAND_DESTROY_CONTEXT;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId = SessionId;

   retCode = Queue_Daemon_Command(&cmd, sizeof(cmd), NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
   if ( reply )
   {
      if ( !reply->Reply.ErrorCode )
      {
         retCode = 0;
      }
      else
      {
         retCode = -EIO;
      }
      Novfs_Free(reply);
   }
   return(retCode);
}

/*++======================================================================*/
int Daemon_Get_UserSpace( uint64_t SessionId, uint64_t *TotalSize, uint64_t *Free, uint64_t *TotalEnties, uint64_t *FreeEnties)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   GET_USER_SPACE_REQUEST cmd;
   PGET_USER_SPACE_REPLY  reply;
   u_long replylen=0;
   int retCode=0;

   DbgPrint("Daemon_Get_UserSpace: 0x%llx\n", SessionId);

   cmd.Command.CommandType = VFS_COMMAND_GET_USER_SPACE;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId = SessionId;

   retCode = Queue_Daemon_Command(&cmd, sizeof(cmd), NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
   if ( reply )
   {
      if ( !reply->Reply.ErrorCode )
      {

         DbgPrint("TotalSpace:  %llu\n", reply->TotalSpace);
         DbgPrint("FreeSpace:   %llu\n", reply->FreeSpace);
         DbgPrint("TotalEnties: %llu\n", reply->TotalEnties);
         DbgPrint("FreeEnties:  %llu\n", reply->FreeEnties);

         if (TotalSize)   *TotalSize   = reply->TotalSpace;
         if (Free)        *Free        = reply->FreeSpace;
         if (TotalEnties) *TotalEnties = reply->TotalEnties;
         if (FreeEnties)  *FreeEnties  = reply->FreeEnties;
         retCode = 0;
      }
      else
      {
         retCode = -EIO;
      }
      Novfs_Free(reply);
   }
   return(retCode);
}

/*++======================================================================*/
int Daemon_SetMountPoint ( char *Path )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   PSET_MOUNT_PATH_REQUEST cmd;
   PSET_MOUNT_PATH_REPLY  reply;
   u_long replylen, cmdlen;
   int retCode = -ENOMEM;

   DbgPrint("Daemon_SetMountPoint: %s\n", Path);
   
   replylen = strlen(Path);

   cmdlen = sizeof(SET_MOUNT_PATH_REQUEST) + replylen;

   cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
   if ( cmd )
   {
      cmd->Command.CommandType = VFS_COMMAND_SET_MOUNT_PATH;
      cmd->Command.SequenceNumber = 0;
      cmd->Command.SessionId = 0;
      cmd->PathLength = replylen;

      strcpy(cmd->Path, Path);

      replylen = 0;

      retCode = Queue_Daemon_Command(cmd, cmdlen, NULL, 0, (void *)&reply, &replylen, INTERRUPTIBLE);
      if ( reply )
      {
         if ( !reply->Reply.ErrorCode )
         {
            retCode = 0;
         }
         else
         {
            retCode = -EIO;
         }
         Novfs_Free(reply);
      }
      Novfs_Free(cmd);
   }
   return(retCode);
}

/*++======================================================================*/
int Daemon_SendDebugCmd ( char *Command )
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   DEBUG_REQUEST cmd;
   PDEBUG_REPLY  reply;
   DEBUG_REPLY   lreply;
   u_long replylen, cmdlen;
   DATA_LIST dlist[2];

   int retCode = -ENOMEM;

   DbgPrint("Daemon_SendDebugCmd: %s\n", Command);
   
   dlist[0].page   = NULL;
   dlist[0].offset = (char *)Command;
   dlist[0].len    = strlen(Command);
   dlist[0].rwflag = DLREAD;

   dlist[1].page   = NULL;
   dlist[1].offset = (char *)&lreply;
   dlist[1].len    = sizeof(lreply);
   dlist[1].rwflag = DLWRITE;

   cmdlen = (int)(&((PDEBUG_REQUEST)0)->dbgcmd);

   cmd.Command.CommandType = VFS_COMMAND_DBG;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId = 0;
   cmd.cmdlen = strlen(Command);

   replylen = 0;

   retCode = Queue_Daemon_Command(&cmd, cmdlen, dlist, 2, (void *)&reply, &replylen, INTERRUPTIBLE);
   if ( reply )
   {
      Novfs_Free(reply);
   }
   if (0 == retCode)
   {
      retCode = lreply.Reply.ErrorCode;
   }

   return(retCode);
}

/*++======================================================================*/
int 
NO_TRACE
Daemon_ioctl(struct inode *inode, struct file *file, unsigned int cmd, u_long arg)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -ENOSYS;

   switch (cmd)
   {
      case IOC_LOGIN:
      {
         retCode = Daemon_Login((PLOGIN)arg, Scope_Get_SessionId( NULL ));
         break;
      }

      case IOC_LOGOUT:
      {
         retCode = Daemon_Logout((PLOGOUT)arg, Scope_Get_SessionId( NULL ));
         break;
      }
      case IOC_DEBUGPRINT:
      {
         struct Ioctl_Debug {
            int length;
            char *data;
         } io;
         char *buf;
         io.length = 0;
         copy_from_user(&io, (char *)arg, sizeof(io));
         if (io.length)
         {
            buf = Novfs_Malloc(io.length+1, GFP_KERNEL);
            if (buf)
            {
               buf[0] = 0;
               copy_from_user(buf, io.data, io.length);
               buf[io.length] = '\0';
               DbgPrint("%s", buf);
               Novfs_Free(buf);
               retCode = 0;
            }
         }
         break;
      }

      case IOC_XPLAT:
      {
         XPLAT   data;

         copy_from_user(&data, (void *)arg, sizeof(data));
         retCode = ((data.xfunction & 0x0000FFFF) | 0xCC000000);

         switch(data.xfunction)
         {
            case NWC_GET_MOUNT_PATH:
               DbgPrint("[Daemon_ioctl] Call NwdGetMountPath\n");
               retCode = NwdGetMountPath( &data );
               break;
         }

         DbgPrint("[NOVFS XPLAT] status Code = %X\n", retCode);
         break;
      }

   }
   return (retCode);
}

/*++======================================================================*/
int Daemon_Added_Resource(daemon_handle_t *DHandle, int Type, u_long CHandle, u_char *FHandle, u_long Mode, u_long Size)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_resource_t *resource;
   int retVal = NWE_OUT_OF_HEAP_SPACE;

   if (FHandle)
      DbgPrint("Daemon_Added_Resource: DHandle=0x%p Type=%d CHandle=0x%x FHandle=0x%x Mode=0x%x Size=%d\n", DHandle, Type, CHandle, *(u_long *)&FHandle[2], Mode, Size);
   else
      DbgPrint("Daemon_Added_Resource: DHandle=0x%p Type=%d CHandle=0x%x\n", DHandle, Type, CHandle);

   resource = Novfs_Malloc(sizeof(daemon_resource_t), GFP_KERNEL);
   if (resource)
   {
      resource->type = Type;
      resource->connection = CHandle;
      if (FHandle)
      {
         memcpy( resource->handle, FHandle, sizeof(resource->handle) );
      }
      else
      {
         memset( resource->handle, 0, sizeof(resource->handle) );
      }
      resource->mode = Mode; 
      resource->size = Size;
      write_lock( &DHandle->lock );
      list_add( &resource->list, &DHandle->list );
      write_unlock( &DHandle->lock );
      DbgPrint("Daemon_Added_Resource: Adding resource=0x%p\n", resource);
      retVal = 0;
   }

   return(retVal);
}

/*++======================================================================*/
int Daemon_Remove_Resource(daemon_handle_t *DHandle, int Type, u_long CHandle, u_long FHandle)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_resource_t *resource;
   struct list_head *l;
   int retVal = -ENOMEM;

   DbgPrint("Daemon_Remove_Resource: DHandle=0x%p Type=%d CHandle=0x%x FHandle=0x%x\n", DHandle, Type, CHandle, FHandle);

   write_lock( &DHandle->lock );
      
   list_for_each( l, &DHandle->list )
   {
      resource = list_entry( l, daemon_resource_t, list );
      
      if ( (Type == resource->type) &&
           (resource->connection == CHandle) )
      {
         DbgPrint("Daemon_Remove_Resource: Found resource=0x%p\n", resource);
         l = l->prev;
         list_del( &resource->list );
         Novfs_Free( resource );
         break;
      }
   }

   write_unlock( &DHandle->lock );

   return(retVal);
}

/*++======================================================================*/
int Daemon_Library_open(struct inode *inode, struct file *file)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retVal = -ENOMEM;
   daemon_handle_t *dh;

   DbgPrint("Daemon_Library_open: inode=0x%p file=0x%p\n", inode, file);

   if ((dh = Novfs_Malloc(sizeof(daemon_handle_t), GFP_KERNEL)))
   {
      file->private_data = dh;
      INIT_LIST_HEAD( &dh->list);
      rwlock_init( &dh->lock );
      dh->session = Scope_Get_SessionId( NULL );
      retVal = 0;
   }
   return(retVal); 
}

/*++======================================================================*/
int Daemon_Library_close(struct inode *inode, struct file *file)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_handle_t *dh;
   daemon_resource_t *resource;
   struct list_head *l;

   char commanddata[sizeof(XPLAT_CALL_REQUEST)+sizeof(NwdCCloseConn)];
   PXPLAT_CALL_REQUEST  cmd;
   PXPLAT_CALL_REPLY		reply;
   PNwdCCloseConn			nwdClose;
   u_long			cmdlen, replylen;

   DbgPrint("Daemon_Library_close: inode=0x%p file=0x%p\n", inode, file);
   if (file->private_data)
   {
      dh = (daemon_handle_t *)file->private_data;

      list_for_each( l, &dh->list )
      {
         resource = list_entry( l, daemon_resource_t, list );
         
         if (DH_TYPE_STREAM == resource->type)
         {
            Novfs_Close_Stream( resource->connection, resource->handle, dh->session );
         }
         else if (DH_TYPE_CONNECTION == resource->type)
         {
	         cmd = (PXPLAT_CALL_REQUEST)commanddata;
	         cmdlen = offsetof(XPLAT_CALL_REQUEST, data) + sizeof(NwdCCloseConn);
		      cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		      cmd->Command.SequenceNumber = 0;
		      cmd->Command.SessionId      = dh->session;
		      cmd->NwcCommand = NWC_CLOSE_CONN;

		      cmd->dataLen = sizeof(NwdCCloseConn);
		      nwdClose = (PNwdCCloseConn)cmd->data;
		      nwdClose->ConnHandle = resource->connection;

		      Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, 0);
		      if (reply)
		      {
			      Novfs_Free(reply);
		      }
         }
         l = l->prev;
         list_del( &resource->list );
         Novfs_Free( resource );
      }
      Novfs_Free(dh);
      file->private_data = NULL;
	}

   return(0);
}

/*++======================================================================*/
ssize_t Daemon_Library_read(struct file *file, char *buf, size_t len, loff_t *off)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_handle_t *dh;
   daemon_resource_t *resource;

   size_t thisread, totalread=0;
   loff_t offset = *off;

   DbgPrint("Daemon_Library_read: file=0x%p len=%d off=%lld\n",  file, len, *off);

   if (file->private_data)
   {
      dh = file->private_data;
      read_lock( &dh->lock );
      if (&dh->list != dh->list.next)
      {
         resource = list_entry( dh->list.next, daemon_resource_t, list );
         
         if (DH_TYPE_STREAM == resource->type)
         {
            while( len > 0 && (offset < resource->size) )
            {
               thisread = len;
               if (Novfs_Read_Stream(resource->connection, resource->handle, buf, &thisread, &offset, 1, dh->session) || !thisread)
               {
                  break;
               }
               len -= thisread;
               buf += thisread;
               offset += thisread;
               totalread += thisread;
            }
         }
      }
      read_unlock( &dh->lock );
   }
   *off = offset;
   DbgPrint("Daemon_Library_read return = 0x%x\n", totalread);
   return(totalread);
}

/*++======================================================================*/
ssize_t Daemon_Library_write(struct file *file, const char *buf, size_t len, loff_t *off)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_handle_t *dh;
   daemon_resource_t *resource;

   size_t thiswrite, totalwrite=0;
   loff_t offset = *off;
   int status;

   DbgPrint("Daemon_Library_write: file=0x%p len=%d off=%lld\n",  file, len, *off);

   if (file->private_data)
   {
      dh = file->private_data;
      read_lock( &dh->lock );
      if (&dh->list != dh->list.next)
      {
         resource = list_entry( dh->list.next, daemon_resource_t, list );
         
         if (DH_TYPE_STREAM == resource->type)
         {
            while(len > 0)
            {
               thiswrite = len;
               status = Novfs_Write_Stream(resource->connection, resource->handle, (PVOID)buf, &thiswrite, &offset, dh->session);
               if (status || !thiswrite)
               {
                  totalwrite = status;
                  break;
               }
               len -= thiswrite;
               buf += thiswrite;
               offset += thiswrite;
               totalwrite += thiswrite;
               if (offset > resource->size)
               {
                  resource->size = offset;
               }
            }
         }
      }
      read_unlock( &dh->lock );
   }
   *off = offset;
   DbgPrint("Daemon_Library_write return = 0x%x\n", totalwrite);

   return(totalwrite);
}

/*++======================================================================*/
int 
NO_TRACE
Daemon_Library_ioctl(struct inode *inode, struct file *file, unsigned int cmd, u_long arg)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -ENOSYS;
   daemon_handle_t *dh;
   u_long handle=0;


   dh = file->private_data;

   DbgPrint("Daemon_Library_ioctl: file=0x%p 0x%x 0x%x dh=0x%p\n", file, cmd, arg, dh);

   if (dh)
   {

      switch (cmd)
      {
         case IOC_LOGIN:
         {
            retCode = Daemon_Login((PLOGIN)arg, dh->session);
            break;
         }

         case IOC_LOGOUT:
         {
            retCode = Daemon_Logout((PLOGOUT)arg, dh->session);
            break;
         }

         case IOC_DEBUGPRINT:
         {
            struct Ioctl_Debug {
               int length;
               char *data;
            } io;
            char *buf;
            io.length = 0;
            copy_from_user(&io, (char *)arg, sizeof(io));
            if (io.length)
            {
               buf = Novfs_Malloc(io.length+1, GFP_KERNEL);
               if (buf)
               {
                  buf[0] = 0;
                  copy_from_user(buf, io.data, io.length);
                  buf[io.length] = '\0';
                  DbgPrint("%s", buf);
                  Novfs_Free(buf);
                  retCode = 0;
               }
            }
            break;
         }

         case IOC_XPLAT:
         {
            XPLAT   data;

            copy_from_user(&data, (void *)arg, sizeof(data));
            retCode = ((data.xfunction & 0x0000FFFF) | 0xCC000000);

            switch(data.xfunction)
            {
               case NWC_OPEN_CONN_BY_NAME:
                  DbgPrint("[VFS XPLAT] Call NwOpenConnByName\n");
                  retCode = NwOpenConnByName(&data, &handle, dh->session);
                  if ( !retCode )
                  {
                     Daemon_Added_Resource(dh, DH_TYPE_CONNECTION, handle, 0, 0, 0);
                  }
                  break;

               case NWC_OPEN_CONN_BY_ADDRESS:
                  DbgPrint("[VFS XPLAT] Call NwOpenConnByAddress\n");
                  retCode = NwOpenConnByAddr(&data, &handle, dh->session);
                  if ( !retCode )
                  {
                     Daemon_Added_Resource(dh, DH_TYPE_CONNECTION, handle, 0, 0, 0);
                  }
                  break;

               case NWC_OPEN_CONN_BY_REFERENCE:

                  DbgPrint("[VFS XPLAT] Call NwOpenConnByReference\n");
                  retCode = NwOpenConnByRef(&data, &handle, dh->session);
                  if ( !retCode )
                  {
                     Daemon_Added_Resource(dh, DH_TYPE_CONNECTION, handle, 0, 0, 0);
                  }
                  break;

               case NWC_SYS_CLOSE_CONN:
                  DbgPrint("[VFS XPLAT] Call NwSysCloseConn\n");
                  retCode = NwSysConnClose(&data, &handle, dh->session);
                  Daemon_Remove_Resource(dh, DH_TYPE_CONNECTION, handle, 0);
                  break;

               case NWC_CLOSE_CONN:
                  DbgPrint("[VFS XPLAT] Call NwCloseConn\n");
                  retCode = NwConnClose(&data, &handle, dh->session);
                  Daemon_Remove_Resource(dh, DH_TYPE_CONNECTION, handle, 0);
                  break;

               case NWC_LOGIN_IDENTITY:
                  DbgPrint("[VFS XPLAT] Call NwLoginIdentity\n");
                  retCode = NwLoginIdentity(&data, dh->session);
                  break;

               case NWC_RAW_NCP_REQUEST:
                  DbgPrint("[VFS XPLAT] Send Raw NCP Request\n");
                  retCode = NwRawSend(&data, dh->session);
                  break;

               case NWC_AUTHENTICATE_CONN_WITH_ID:
                  DbgPrint("[VFS XPLAT] Authenticate Conn With ID\n");
                  retCode = NwAuthConnWithId(&data, dh->session);
                  break;

			      case NWC_UNAUTHENTICATE_CONN:
				      DbgPrint("[VFS XPLAT] UnAuthenticate Conn With ID\n");
				      retCode = NwUnAuthenticate(&data, dh->session);
   			      break;

               case NWC_LICENSE_CONN:
                  DbgPrint("Call NwLicenseConn\n");
                  retCode = NwLicenseConn(&data, dh->session);
                  break;

               case NWC_LOGOUT_IDENTITY:
                  DbgPrint("[VFS XPLAT] Call NwLogoutIdentity\n");
                  retCode = NwLogoutIdentity(&data, dh->session);
                  break;

               case NWC_UNLICENSE_CONN:
                  DbgPrint("[VFS XPLAT] Call NwUnlicense\n");
                  retCode = NwUnlicenseConn(&data, dh->session);
                  break;

               case NWC_GET_CONN_INFO:
                  DbgPrint("[VFS XPLAT] Call NwGetConnInfo\n");
                  retCode = NwGetConnInfo(&data, dh->session);
                  break;

               case NWC_SET_CONN_INFO:
                  DbgPrint("[VFS XPLAT] Call NwGetConnInfo\n");
                  retCode = NwSetConnInfo(&data, dh->session);
                  break;

               case NWC_SCAN_CONN_INFO:
                  DbgPrint("[VFS XPLAT] Call NwScanConnInfo\n");
                  retCode = NwScanConnInfo(&data, dh->session);
                  break;


               case NWC_GET_IDENTITY_INFO:
                  DbgPrint("[VFS XPLAT] Call NwGetIdentityInfo\n");
                  retCode = NwGetIdentityInfo(&data, dh->session);				
                  break;

               case NWC_GET_REQUESTER_VERSION:
                  DbgPrint("[VFS XPLAT] Call NwGetDaemonVersion\n");
                  retCode = NwGetDaemonVersion(&data, dh->session);				
                  break;

               case NWC_GET_PREFERRED_DS_TREE:
                  DbgPrint("[VFS XPLAT] Call NwcGetPreferredDsTree\n");
                  retCode = NwcGetPreferredDSTree(&data, dh->session);
                  break;

               case NWC_SET_PREFERRED_DS_TREE:
                  DbgPrint("[VFS XPLAT] Call NwcSetPreferredDsTree\n");
                  retCode = NwcSetPreferredDSTree(&data, dh->session);
                  break;

               case NWC_GET_DEFAULT_NAME_CONTEXT:
                  DbgPrint("[VFS XPLAT] Call NwcGetDefaultNameContext\n");
                  retCode = NwcGetDefaultNameCtx(&data, dh->session);
                  break;

               case NWC_SET_DEFAULT_NAME_CONTEXT:
                  DbgPrint("[VFS XPLAT] Call NwcSetDefaultNameContext\n");
                  retCode = NwcSetDefaultNameCtx(&data, dh->session);
                  break;

               case NWC_QUERY_FEATURE:
                  DbgPrint("[VFS XPLAT] Call NwQueryFeature\n");
                  retCode = NwQueryFeature(&data, dh->session);
                  break;


               case NWC_GET_TREE_MONITORED_CONN_REF:
                  DbgPrint("[VFS XPLAT] Call NwcGetTreeMonitoredConn\n");
                  retCode = NwcGetTreeMonitoredConn(&data, dh->session);
                  break;

               case NWC_ENUMERATE_IDENTITIES:
                  DbgPrint("[VFS XPLAT] Call NwcEnumerateIdentities\n");
                  retCode = NwcEnumIdentities(&data, dh->session);
                  break;

               case NWC_CHANGE_KEY:
                  DbgPrint("[VFS XPLAT] Call NwcChangeAuthKey\n");
                  retCode = NwcChangeAuthKey(&data, dh->session);
                  break;

               case NWC_CONVERT_LOCAL_HANDLE:
                  DbgPrint("[VFS XPLAT] Call NwdConvertLocalHandle\n");
                  retCode = NwdConvertLocalHandle(&data, dh);
                  break;
                  
               case NWC_CONVERT_NETWARE_HANDLE:
                  DbgPrint("[VFS XPLAT] Call NwdConvertNetwareHandle\n");
                  retCode = NwdConvertNetwareHandle(&data, dh);
                  break;

               case NWC_SET_PRIMARY_CONN:
                  DbgPrint("[VFS XPLAT] Call NwcSetPrimaryConn\n");
                  retCode = NwcSetPrimaryConn(&data, dh->session);
                  break;

               case NWC_GET_PRIMARY_CONN:
                  DbgPrint("[VFS XPLAT] Call NwcGetPrimaryConn\n");
                  retCode = NwcGetPrimaryConn(&data, dh->session);
                  break;

               case NWC_MAP_DRIVE:
                  DbgPrint("[VFS XPLAT] Call NwcMapDrive\n");
                  retCode = NwcSetMapDrive(&data, dh->session);
                  break;

               case NWC_UNMAP_DRIVE:
                  DbgPrint("[VFS XPLAT] Call NwcUnMapDrive\n");
                  retCode = NwcUnMapDrive(&data, dh->session);
                  break;

               case NWC_ENUMERATE_DRIVES:
                  DbgPrint("[VFS XPLAT] Call NwcEnumerateDrives\n");
                  retCode = NwcEnumerateDrives(&data, dh->session);
                  break;

               case NWC_GET_MOUNT_PATH:
                  DbgPrint("[VFS XPLAT] Call NwdGetMountPath\n");
                  retCode = NwdGetMountPath( &data );
                  break;

               case NWC_GET_BROADCAST_MESSAGE:
                  DbgPrint("[VSF XPLAT Call NwdGetBroadcastMessage\n");
                  retCode = NwcGetBroadcastMessage(&data, dh->session);
                  break;

               case NWC_SET_KEY:
                  DbgPrint("[VSF XPLAT Call NwdSetKey\n");
                  retCode = NwdSetKeyValue(&data, dh->session);
                  break;

               case NWC_VERIFY_KEY:
                  DbgPrint("[VSF XPLAT Call NwdVerifyKey\n");
                  retCode = NwdVerifyKeyValue(&data, dh->session);
                  break;

               case NWC_RAW_NCP_REQUEST_ALL:
               case NWC_NDS_RESOLVE_NAME_TO_ID:
               case NWC_FRAGMENT_REQUEST:
               case NWC_GET_CONFIGURED_NSPS:
               default:
               break;

            }

            DbgPrint("[NOVFS XPLAT] status Code = %X\n", retCode);
            break;
         }
      }
   }

   return( retCode );
}

/*++======================================================================*/
unsigned int Daemon_Poll(struct file *file, struct poll_table_struct *poll_table)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   daemon_command_t *que;
   unsigned int mask = POLLOUT | POLLWRNORM;
   
   que = get_next_queue(0);
   if (que)
   {
      mask |= (POLLIN | POLLRDNORM);
   }
   else
   {

   }
   return(mask);
}


void Daemon_Dumpque( void )
{
#ifdef	CONFIG_KDB
   daemon_command_t *que;

   que = (daemon_command_t *)Daemon_Queue.list.next;

   while (que && (que != (daemon_command_t *)&Daemon_Queue.list.next))
   {
      kdb_printf("DaemonQue:\n" \
                 "   Que:       0x%p\n" \
                 "   status:    0x%lx\n" \
                 "   flags:     0x%lx\n" \
                 "   semaphore: 0x%x\n" \
                 "   sequence:  0x%lx\n" \
                 "   timer:     0x%lx\n" \
                 "   request:   0x%p\n" \
                 "   reqlen:    %ld\n" \
                 "   data:      0x%p\n" \
                 "   datalen:   %d\n" \
                 "   reply:     0x%p\n" \
                 "   replen:    %ld\n",
         que,
         que->status,
         que->flags,
         atomic_read(&que->semaphore.count),
         que->sequence,
         que->timer.expires,
         que->request,
         que->reqlen,
         que->data,
         que->datalen,
         que->reply,
         que->replen);
      que = (daemon_command_t *)que->list.next;
   }
#endif
}

/*++======================================================================*/
int NwdConvertNetwareHandle(PXPLAT pdata, daemon_handle_t *DHandle)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retVal;
   NwcConvertNetWareHandle nh;

   DbgPrint("NwdConvertNetwareHandle: DHandle=0x%p\n", DHandle);

	copy_from_user(&nh, pdata->reqData, sizeof(NwcConvertNetWareHandle));

   retVal = Daemon_Added_Resource(DHandle, DH_TYPE_STREAM, nh.ConnHandle, nh.NetWareHandle, nh.uAccessMode, nh.uFileSize);

   return(retVal);
}

/*++======================================================================*/
int NwdConvertLocalHandle(PXPLAT pdata, daemon_handle_t *DHandle)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retVal = NWE_REQUESTER_FAILURE;
   daemon_resource_t *resource;
   NwcConvertLocalHandle lh;
   struct list_head *l;

   DbgPrint("NwdConvertLocalHandle: DHandle=0x%p\n", DHandle);

   read_lock( &DHandle->lock );
      
   list_for_each( l, &DHandle->list )
   {
      resource = list_entry( l, daemon_resource_t, list );
      
      if ( DH_TYPE_STREAM == resource->type )
      {
         lh.uConnReference = resource->connection;

         memcpy(lh.NwFileHandle, resource->handle, sizeof(resource->handle));
         if (pdata->repLen >= sizeof(NwcConvertLocalHandle))
         {
            copy_to_user(pdata->repData, &lh, sizeof(NwcConvertLocalHandle));
            retVal = 0;
         }
         else
         {
            retVal = NWE_BUFFER_OVERFLOW;
         }
            break;
      }
   }

   read_unlock( &DHandle->lock );

   return(retVal);
}

/*++======================================================================*/
int NwdGetMountPath(PXPLAT pdata)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   int retVal = NWE_REQUESTER_FAILURE;
   int len;
   NwcGetMountPath mp;

   copy_from_user(&mp, pdata->reqData, pdata->reqLen);

   if ( Novfs_CurrentMount )
   {

      len = strlen(Novfs_CurrentMount)+1;
      if ( (len > mp.MountPathLen) && mp.pMountPath)
      {
         retVal = NWE_BUFFER_OVERFLOW;
      }
      else
      {
         if (mp.pMountPath)
         {
            copy_to_user(mp.pMountPath, Novfs_CurrentMount, len);
         }
         retVal = 0;
      }

      mp.MountPathLen = len;

      if (pdata->repData && (pdata->repLen >= sizeof(mp)) )
      {
         copy_to_user(pdata->repData, &mp, sizeof(mp));
      }
   }

   return(retVal);
}
